#include "SiriSSL.h"

/* Workaround crummy standard SSL API method names */
#define SSL_peer_cert           SSL_get_peer_certificate
#define SSL_peer_cert_chain     SSL_get_peer_cert_chain
#define SSL_cipher              SSL_get_current_cipher
#define SSL_shared_ciphers      SSL_get_shared_ciphers

/* Our extensions to the SSL API.

   - Additional underlines in function definition
     to show these functions are not part of the standard SSL API (also
     so we don't conflict with any future API functions).

   - We define the macros to allow for easy translation by the XS compiler.
     (avoiding redefinition of PREFIX)
*/

#define SSL_error       SSL__error
#define SSL_clearerr    SSL__clearerr
#define SSL_eof         SSL__eof
#define SSL_close       SSL__close
#define SSL_fdopen      SSL__fdopen
#define SSL_getline     SSL__getline
#define SSL_getlines    SSL__getlines

static int
SSL__error(SSL *ssl)
{
        int ssl_error;
        int io_retval = SSL_get_io_retval(ssl);

        ssl_error = SSL_get_error(ssl, io_retval);

        /* working around an OpenSSL bug */
        if(SSL_get_fd(ssl) < 0 || (ssl_error == SSL_ERROR_SYSCALL
                                   && io_retval == 0))
                return SSL_ERROR_CLOSED;

        return ssl_error;
}

static int
SSL__clearerr(SSL *ssl)
{
        /* kludgy ... */
        SSL_set_io_retval(ssl, 1);
}

static int
SSL__eof(SSL *ssl)
{
        int error = SSL_error(ssl);

        switch(error) {
        case SSL_ERROR_SYSCALL :
        case SSL_ERROR_SSL :
        case SSL_ERROR_CLOSED :
                return error;
        }
        return 0;
}

static void
SSL__close(SSL *ssl)
{
        SSL_shutdown(ssl);
        SSL_set_bio(ssl, NULL, NULL);
}

static int
SSL__fdopen(SSL *ssl, PerlIO *fh)
{
        if(SSL_get_fd(ssl) >= 0)
                SSL_close(ssl);

        SSL_clear(ssl);
        SSL_set_fd(ssl, PerlIO_fileno(fh));
        return SSL_do_handshake(ssl);
}

static SV *
SSL__getline(SSL *ssl)
{
        int n;
        char buf[512], *cp;

        SV *sv_buf = newSVpv("", 0);

        while((n = SSL_peek(ssl, buf, sizeof(buf))) > 0)
        {
                cp = strchr(buf, (int) '\n');
                if(cp != NULL) {
                        int len = (cp - buf) + 1;

                        buf[len] = '\0';
                        sv_catpvn(sv_buf, buf, len);
                        SSL_read(ssl, buf, len);

                        return sv_buf;
                } else {
                        sv_catpvn(sv_buf, buf, n);
                        SSL_read(ssl, buf, n);
                }
        }

        if(SvCUR(sv_buf))
                return sv_buf;
        else {
                SvREFCNT_dec(sv_buf);
                return &sv_undef;
        }
}

MODULE = SiriSSL::SSL   PACKAGE = SiriSSL::SSL  PREFIX = SSL_

SSL *
SSL_new(class, ctx, server, fh)
                SV *            class
                SSL_CTX *       ctx
                int             server
                InOutStream     fh
        PREINIT:
                SSL     *ssl;

        CODE:
                ssl = SSL_new(ctx);

                if(server) SSL_set_accept_state(ssl);
                else SSL_set_connect_state(ssl);

                SSL_set_fd(ssl, PerlIO_fileno(fh));
                if(SSL_do_handshake(ssl) <= 0) {
                        RETVAL = (SSL *)NULL;
                        SSL_free(ssl);
                } else        
                        RETVAL = ssl;

        OUTPUT:
                RETVAL
        
void
SSL_DESTROY(...)
        PREINIT:
                SSL *ssl;
        CODE:
                CLASS_IN(ST(0), ssl, SSL *, SiriSSL::SSL);
                printd("destroying SSL object (0x%x): %d references\n",
                       (unsigned long)ssl, REF_CNT(ssl));
                if(REF_ISLAST(ssl))
                        SSL_close(ssl);
                SSL_free(ssl);

SV *
SSL_fdopen(ssl, fh)
                SSL *           ssl
                InOutStream     fh
        PREINIT:
                int retval;                
        CODE:
                retval = SSL_fdopen(ssl, fh);
                if(retval <= 0)
                        RETVAL = &sv_undef;
                else
                        RETVAL = newSViv(retval);
        OUTPUT:
                RETVAL

SV *
SSL_sysread(ssl, sv_buf, count, offset = 0)
                SSL *           ssl
                SV *            sv_buf
                int             count
                int             offset
        PREINIT:
                int sv_len, size;
        CODE:
                SvPV_force(sv_buf, sv_len);
                SvGROW(sv_buf, count + offset);

                if(offset > sv_len)
                        memset(SvPVX(sv_buf) + sv_len, 0, offset - sv_len);

                size = SSL_read(ssl, SvPVX(sv_buf) + offset, count);
                if(size >= 0)
                        SvCUR_set(sv_buf, size + offset);
                SSL_set_io_retval(ssl, size);

                RETVAL = size < 0 ? &sv_undef : newSViv(size);
        OUTPUT:
                RETVAL

SV *
SSL_syswrite(ssl, sv_buf, length = 0, offset = 0)
                SSL *   ssl
                SV *    sv_buf
                int     length
                int     offset
        PREINIT:
                int     retval;

                SV *    sv_dup;
                int     sv_len;
                char *  sv_cp;
        CODE:
                if(SvPOK(sv_buf))
                        sv_cp = SvPV(sv_buf, sv_len);
                else {
                        sv_dup = sv_2mortal(newSVsv(sv_buf));
                        sv_cp = SvPV_force(sv_dup, sv_len);
                }
                if(items > 2) {
                        int max_len;

                        /* enforce gracefull limits */
                        if(offset > sv_len)
                                offset = sv_len;
                        if(offset < -sv_len)
                                offset = -sv_len;
                        if(offset < 0)
                                offset = sv_len + offset;

                        max_len = sv_len - offset;
                        if(length > max_len)
                                length = max_len;
                } else
                        length = sv_len;

                retval = SSL_write(ssl, sv_cp + offset, length);
                SSL_set_io_retval(ssl, retval);

                RETVAL = retval < 0 ? &sv_undef : newSViv(retval);
        OUTPUT:
                RETVAL
        

int
SSL_peek(ssl, sv_buf, count, offset = 0)
                SSL *   ssl
                SV *    sv_buf
                int     count
                int     offset
        PREINIT:
                int     sv_len;
        CODE:
                SvPV_force(sv_buf, sv_len);
                if(offset > sv_len)
                        offset = sv_len;

                SvGROW(sv_buf, count + offset);
                RETVAL = SSL_peek(ssl, SvPVX(sv_buf) + offset, count);
                SvCUR_set(sv_buf, RETVAL + offset);

        OUTPUT:
                RETVAL

int
SSL_pending(ssl)
        SSL *ssl

X509 *
SSL_peer_cert(ssl)
        SSL *ssl

void
SSL_peer_cert_chain(ssl)
                SSL *ssl
        PREINIT:
                STACK * chain;
                int     i;
        PPCODE:
                if(chain = SSL_peer_cert_chain(ssl))
                        for(i = 0; i < sk_num(chain); i++) {
                                X509 *cert = (X509 *)sk_value(chain, i);
        
                                REF_INC(cert, SSL);
                                XPUSH_OBJ(cert, SiriSSL::X509);
                        }

SSL_CTX *
SSL_ctx(ssl)
                SSL *ssl
        CODE:
                RETVAL = SSL_get_SSL_CTX(ssl);

                /* OpenSSL isn't consistent with references, so this
                   needs to be done here */
                REF_INC(RETVAL, SSL_CTX);
        OUTPUT:
                RETVAL


int
SSL_error(ssl)
        SSL *ssl

int
SSL_clearerr(ssl)
        SSL *ssl

int
SSL_eof(ssl)
        SSL *ssl

SV *
SSL_shared_ciphers(ssl)
                SSL *   ssl
        PREINIT:
                SV *    sv_buf;
                char *  sv_cp;
        CODE:
                SV_ALLOC(sv_buf, CIPHERS_MAX_LEN);
                sv_cp = SvPVX(sv_buf);

                SSL_get_shared_ciphers(ssl, sv_cp, CIPHERS_MAX_LEN);
                SvCUR_set(sv_buf, strlen(sv_cp));

                RETVAL = sv_buf;
        OUTPUT:
                RETVAL

int
SSL_fileno(ssl)
        SSL *ssl
        CODE:
                RETVAL = SSL_get_fd(ssl);
        OUTPUT:
                RETVAL

void
SSL_close(ssl)
        SSL *ssl


int
SSL_renegotiate(ssl)
        SSL *ssl

int
SSL_get_verify_result(ssl)
        SSL *ssl

SV *
SSL_getline(ssl)
        SSL *ssl

SSL_CIPHER *
SSL_cipher(ssl)
        SSL *ssl

void
SSL_getlines(ssl)
                SSL *ssl
        PREINIT:
                SV *line;
        PPCODE:
                while((line = SSL_getline(ssl)) != &sv_undef)
                        XPUSHs(sv_2mortal(line));
